@@ -9,6 +9,7 @@
  * daniel@veillard.com
  */
 
+#define IN_LIBXSLT
 #include "libxslt.h"
 
 #include <string.h>
@@ -74,11 +75,14 @@
  * an attribute set
  */
 
+
 typedef struct _xsltAttrElem xsltAttrElem;
 typedef xsltAttrElem *xsltAttrElemPtr;
 struct _xsltAttrElem {
     struct _xsltAttrElem *next;/* chained list */
     xmlNodePtr attr;	/* the xsl:attribute definition */
+    const xmlChar *set; /* or the attribute set */
+    const xmlChar *ns;  /* and its namespace */
 };
 
 /************************************************************************
@@ -118,7 +122,10 @@ xsltNewAttrElem(xmlNodePtr attr) {
  */
 static void
 xsltFreeAttrElem(xsltAttrElemPtr attr) {
-    memset(attr, -1, sizeof(xsltAttrElem));
+    if (attr->set != NULL)
+	xmlFree((char *)attr->set);
+    if (attr->ns != NULL)
+	xmlFree((char *)attr->ns);
     xmlFree(attr);
 }
 
@@ -185,38 +192,64 @@ xsltMergeAttrElemList(xsltAttrElemPtr list, xsltAttrElemPtr old) {
     int add;
 
     while (old != NULL) {
-
+	if ((old->attr == NULL) && (old->set == NULL)) {
+	    old = old->next;
+	    continue;
+	}
 	/*
 	 * Check that the attribute is not yet in the list
 	 */
 	cur = list;
 	add = 1;
 	while (cur != NULL) {
+	    if ((cur->attr == NULL) && (cur->set == NULL)) {
+		if (cur->next == NULL)
+		    break;
+		cur = cur->next;
+		continue;
+	    }
+	    if ((cur->set != NULL) && (cur->set == old->set)) {
+		add = 0;
+		break;
+	    }
+	    if (cur->set != NULL) {
+		if (cur->next == NULL)
+		    break;
+		cur = cur->next;
+		continue;
+	    }
+	    if (old->set != NULL) {
+		if (cur->next == NULL)
+		    break;
+		cur = cur->next;
+		continue;
+	    }
 	    if (cur->attr == old->attr) {
 		xsltGenericError(xsltGenericErrorContext,
 	     "xsl:attribute-set : use-attribute-sets recursion detected\n");
 		return(list);
 	    }
-	    if (xmlStrEqual(cur->attr->name, old->attr->name)) {
-		if (cur->attr->ns == old->attr->ns) {
-		    add = 0;
-		    break;
-		}
-		if ((cur->attr->ns != NULL) && (old->attr->ns != NULL) &&
-		    (xmlStrEqual(cur->attr->ns->href, old->attr->ns->href))) {
-		    add = 0;
-		    break;
-		}
-	    }
 	    if (cur->next == NULL)
 		break;
             cur = cur->next;
 	}
 
-	if (cur == NULL) {
-	    list = xsltNewAttrElem(old->attr);
-	} else if (add) {
-	    cur->next = xsltNewAttrElem(old->attr);
+	if (add == 1) {
+	    if (cur == NULL) {
+		list = xsltNewAttrElem(old->attr);
+		if (old->set != NULL) {
+		    list->set = xmlStrdup(old->set);
+		    if (old->ns != NULL)
+			list->ns = xmlStrdup(old->ns);
+		}
+	    } else if (add) {
+		cur->next = xsltNewAttrElem(old->attr);
+		if (old->set != NULL) {
+		    cur->next->set = xmlStrdup(old->set);
+		    if (old->ns != NULL)
+			cur->next->ns = xmlStrdup(old->ns);
+		}
+	    }
 	}
 
 	old = old->next;
@@ -335,8 +368,12 @@ xsltParseStylesheetAttributeSet(xsltStylesheetPtr style, xmlNodePtr cur) {
 		attrib = NULL;
 		prefix = NULL;
 	    }
-	    values2 = xmlHashLookup2(style->attributeSets, ncname2, prefix2);
-	    values = xsltMergeAttrElemList(values, values2);
+	    values2 = xsltNewAttrElem(NULL);
+	    if (values2 != NULL) {
+		values2->set = ncname2;
+		values2->ns = prefix2;
+		values = xsltMergeAttrElemList(values, values2);
+	    }
 
 	    if (attrib != NULL)
 		xmlFree(attrib);
@@ -353,6 +390,8 @@ done:
     /*
      * Update the value
      */
+    if (values == NULL)
+	values = xsltNewAttrElem(NULL);
     xmlHashUpdateEntry2(style->attributeSets, ncname, prefix, values, NULL);
 #ifdef WITH_XSLT_DEBUG_ATTRIBUTES
     xsltGenericDebug(xsltGenericDebugContext,
@@ -368,6 +407,105 @@ error:
         xmlFree(prefix);
 }
 
+/**
+ * xsltGetSAS:
+ * @style:  the XSLT stylesheet
+ * @name:  the attribute list name
+ * @ns:  the attribute list namespace
+ *
+ * lookup an attribute set based on the style cascade
+ *
+ * Returns the attribute set or NULL
+ */
+static xsltAttrElemPtr
+xsltGetSAS(xsltStylesheetPtr style, const xmlChar *name, const xmlChar *ns) {
+    xsltAttrElemPtr values;
+
+    while (style != NULL) {
+	values = xmlHashLookup2(style->attributeSets, name, ns);
+	if (values != NULL)
+	    return(values);
+	style = xsltNextImport(style);
+    }
+    return(NULL);
+}
+
+/**
+ * xsltResolveSASCallback,:
+ * @style:  the XSLT stylesheet
+ *
+ * resolve the references in an attribute set.
+ */
+static void
+xsltResolveSASCallback(xsltAttrElemPtr values, xsltStylesheetPtr style,
+	               const xmlChar *name, const xmlChar *ns,
+		       const xmlChar *ignored) {
+    xsltAttrElemPtr tmp;
+    xsltAttrElemPtr refs;
+
+    tmp = values;
+    while (tmp != NULL) {
+	if (tmp->set != NULL) {
+	    /*
+	     * Check against cycles !
+	     */
+	    if ((xmlStrEqual(name, tmp->set)) && (xmlStrEqual(ns, tmp->ns))) {
+		xsltGenericError(xsltGenericErrorContext,
+     "xsl:attribute-set : use-attribute-sets recursion detected on %s\n",
+                                 name);
+	    } else {
+#ifdef WITH_XSLT_DEBUG_ATTRIBUTES
+		xsltGenericDebug(xsltGenericDebugContext,
+			"Importing attribute list %s\n", tmp->set);
+#endif
+
+		refs = xsltGetSAS(style, tmp->set, tmp->ns);
+		if (refs == NULL) {
+		    xsltGenericError(xsltGenericErrorContext,
+     "xsl:attribute-set : use-attribute-sets %s reference missing %s\n",
+				     name, tmp->set);
+		} else {
+		    /*
+		     * recurse first for cleanup
+		     */
+		    xsltResolveSASCallback(refs, style, name, ns, NULL);
+		    /*
+		     * Then merge
+		     */
+		    xsltMergeAttrElemList(values, refs);
+		    /*
+		     * Then suppress the reference
+		     */
+		    xmlFree((char *)tmp->set);
+		    tmp->set = NULL;
+		    if (tmp->ns != NULL) {
+			xmlFree((char *)tmp->ns);
+		    }
+		}
+	    }
+	}
+	tmp = tmp->next;
+    }
+}
+
+/**
+ * xsltResolveStylesheetAttributeSet:
+ * @style:  the XSLT stylesheet
+ *
+ * resolve the references between attribute sets.
+ */
+void
+xsltResolveStylesheetAttributeSet(xsltStylesheetPtr style) {
+#ifdef WITH_XSLT_DEBUG_ATTRIBUTES
+    xsltGenericDebug(xsltGenericDebugContext,
+	    "Resolving attribute sets references\n");
+#endif
+    if (style->attributeSets != NULL) {
+	xmlHashScanFull(style->attributeSets, 
+		(xmlHashScannerFull) xsltResolveSASCallback, style);
+    }
+}
+
 /**
  * xsltAttributeInternal:
  * @ctxt:  a XSLT process context
@@ -380,7 +518,9 @@ error:
  */
 static void
 xsltAttributeInternal(xsltTransformContextPtr ctxt, xmlNodePtr node,
-	      xmlNodePtr inst, xsltStylePreCompPtr comp, int fromset) {
+                      xmlNodePtr inst, xsltStylePreCompPtr comp,
+                      int fromset)
+{
     xmlChar *prop = NULL;
     xmlChar *ncname = NULL, *name, *namespace;
     xmlChar *prefix = NULL;
@@ -391,97 +531,114 @@ xsltAttributeInternal(xsltTransformContextPtr ctxt, xmlNodePtr node,
 
 
     if (ctxt->insert == NULL)
-	return;
+        return;
     if (comp == NULL) {
-	xsltPrintErrorContext(ctxt, NULL, inst);
-	xsltGenericError(xsltGenericErrorContext,
-	     "xsl:attribute : compilation failed\n");
-	return;
+        xsltPrintErrorContext(ctxt, NULL, inst);
+        xsltGenericError(xsltGenericErrorContext,
+                         "xsl:attribute : compilation failed\n");
+        return;
     }
 
-    if ((ctxt == NULL) || (node == NULL) || (inst == NULL) || (comp == NULL))
-	return;
+    if ((ctxt == NULL) || (node == NULL) || (inst == NULL)
+        || (comp == NULL))
+        return;
     if (!comp->has_name) {
-	return;
+        return;
     }
     if (ctxt->insert->children != NULL) {
-	xsltPrintErrorContext(ctxt, NULL, inst);
-	xsltGenericError(xsltGenericErrorContext,
-	     "xsl:attribute : node already has children\n");
-	return;
+        xsltPrintErrorContext(ctxt, NULL, inst);
+        xsltGenericError(xsltGenericErrorContext,
+                         "xsl:attribute : node already has children\n");
+        return;
     }
+#ifdef WITH_DEBUGGER
+    if (xslDebugStatus != XSLT_DEBUG_NONE) {
+        xslHandleDebugger(inst, node, NULL, ctxt);
+    }
+#endif
+
     if (comp->name == NULL) {
-	prop = xsltEvalAttrValueTemplate(ctxt, inst, (const xmlChar *)"name",
-		                         XSLT_NAMESPACE);
-	if (prop == NULL) {
-	    xsltPrintErrorContext(ctxt, NULL, inst);
-	    xsltGenericError(xsltGenericErrorContext,
-		 "xsl:attribute : name is missing\n");
-	    goto error;
-	}
-	name = prop;
+        prop =
+            xsltEvalAttrValueTemplate(ctxt, inst, (const xmlChar *) "name",
+                                      XSLT_NAMESPACE);
+        if (prop == NULL) {
+            xsltPrintErrorContext(ctxt, NULL, inst);
+            xsltGenericError(xsltGenericErrorContext,
+                             "xsl:attribute : name is missing\n");
+            goto error;
+        }
+        name = prop;
     } else {
-	name = comp->name;
+        name = comp->name;
     }
 
     ncname = xmlSplitQName2(name, &prefix);
     if (ncname == NULL) {
-	prefix = NULL;
+        prefix = NULL;
     } else {
-	name = ncname;
+        name = ncname;
     }
-    if (!xmlStrncasecmp(prefix, (xmlChar *)"xml", 3)) {
+    if (!xmlStrncasecmp(prefix, (xmlChar *) "xml", 3)) {
 #ifdef WITH_XSLT_DEBUG_PARSING
-	xsltGenericDebug(xsltGenericDebugContext,
-	     "xsltAttribute: xml prefix forbidden\n");
+        xsltGenericDebug(xsltGenericDebugContext,
+                         "xsltAttribute: xml prefix forbidden\n");
 #endif
-	goto error;
+        goto error;
     }
     if ((comp->ns == NULL) && (comp->has_ns)) {
-	namespace = xsltEvalAttrValueTemplate(ctxt, inst,
-		(const xmlChar *)"namespace", XSLT_NAMESPACE);
-	if (namespace != NULL) {
-	    ns = xsltGetSpecialNamespace(ctxt, inst, namespace, prefix,
-		                         ctxt->insert);
-	    xmlFree(namespace);
-	} else {
-	    if (prefix != NULL) {
-		ns = xmlSearchNs(inst->doc, inst, prefix);
-		if (ns == NULL) {
-		    xsltPrintErrorContext(ctxt, NULL, inst);
-		    xsltGenericError(xsltGenericErrorContext,
-			"xsl:attribute : no namespace bound to prefix %s\n", prefix);
-		} else {
-		    ns = xsltGetNamespace(ctxt, inst, ns, ctxt->insert);
-		}
-	    }
-	}
+        namespace = xsltEvalAttrValueTemplate(ctxt, inst,
+                                              (const xmlChar *)
+                                              "namespace", XSLT_NAMESPACE);
+        if (namespace != NULL) {
+            ns = xsltGetSpecialNamespace(ctxt, inst, namespace, prefix,
+                                         ctxt->insert);
+            xmlFree(namespace);
+        } else {
+            if (prefix != NULL) {
+                ns = xmlSearchNs(inst->doc, inst, prefix);
+                if (ns == NULL) {
+                    xsltPrintErrorContext(ctxt, NULL, inst);
+                    xsltGenericError(xsltGenericErrorContext,
+                                     "xsl:attribute : no namespace bound to prefix %s\n",
+                                     prefix);
+                } else {
+                    ns = xsltGetNamespace(ctxt, inst, ns, ctxt->insert);
+                }
+            }
+        }
     } else if (comp->ns != NULL) {
-	ns = xsltGetSpecialNamespace(ctxt, inst, comp->ns, prefix,
-				     ctxt->insert);
+        ns = xsltGetSpecialNamespace(ctxt, inst, comp->ns, prefix,
+                                     ctxt->insert);
+    } else if (prefix != NULL) {
+	xmlNsPtr tmp;
+	tmp = xmlSearchNs(inst->doc, inst, prefix);
+	if (tmp != NULL) {
+	    ns = xsltGetNamespace(ctxt, inst, tmp, ctxt->insert);
+	}
     }
 
     if ((fromset) && (ns != NULL))
-	URL = ns->href;
+        URL = ns->href;
     if ((fromset == 0) || (!xmlHasNsProp(ctxt->insert, name, URL))) {
-	value = xsltEvalTemplateString(ctxt, node, inst);
-	if (value == NULL) {
-	    if (ns) {
-		attr = xmlSetNsProp(ctxt->insert, ns, name, 
-				    (const xmlChar *)"");
-	    } else {
-		attr = xmlSetProp(ctxt->insert, name, (const xmlChar *)"");
-	    }
-	} else {
-	    if (ns) {
-		attr = xmlSetNsProp(ctxt->insert, ns, name, value);
-	    } else {
-		attr = xmlSetProp(ctxt->insert, name, value);
-	    }
-	}
+        value = xsltEvalTemplateString(ctxt, node, inst);
+        if (value == NULL) {
+            if (ns) {
+                attr = xmlSetNsProp(ctxt->insert, ns, name,
+                                    (const xmlChar *) "");
+            } else {
+                attr =
+                    xmlSetProp(ctxt->insert, name, (const xmlChar *) "");
+            }
+        } else {
+            if (ns) {
+                attr = xmlSetNsProp(ctxt->insert, ns, name, value);
+            } else {
+                attr = xmlSetProp(ctxt->insert, name, value);
+            }
+        }
     }
 
-error:
+  error:
     if (prop != NULL)
         xmlFree(prop);
     if (ncname != NULL)
@@ -519,7 +676,9 @@ xsltAttribute(xsltTransformContextPtr ctxt, xmlNodePtr node,
 
 void
 xsltApplyAttributeSet(xsltTransformContextPtr ctxt, xmlNodePtr node,
-	              xmlNodePtr inst ATTRIBUTE_UNUSED, xmlChar *attributes) {
+                      xmlNodePtr inst ATTRIBUTE_UNUSED,
+                      xmlChar * attributes)
+{
     xmlChar *ncname = NULL;
     xmlChar *prefix = NULL;
     xmlChar *attrib, *endattr;
@@ -527,49 +686,61 @@ xsltApplyAttributeSet(xsltTransformContextPtr ctxt, xmlNodePtr node,
     xsltStylesheetPtr style;
 
     if (attributes == NULL) {
-	return;
+        return;
     }
 
     attrib = attributes;
     while (*attrib != 0) {
-	while (IS_BLANK(*attrib)) attrib++;
-	if (*attrib == 0)
-	    break;
+        while (IS_BLANK(*attrib))
+            attrib++;
+        if (*attrib == 0)
+            break;
         endattr = attrib;
-	while ((*endattr != 0) && (!IS_BLANK(*endattr))) endattr++;
-	attrib = xmlStrndup(attrib, endattr - attrib);
-	if (attrib) {
+        while ((*endattr != 0) && (!IS_BLANK(*endattr)))
+            endattr++;
+        attrib = xmlStrndup(attrib, endattr - attrib);
+        if (attrib) {
 #ifdef WITH_XSLT_DEBUG_ATTRIBUTES
-	    xsltGenericDebug(xsltGenericDebugContext,
-		"apply attribute set %s\n", attrib);
+            xsltGenericDebug(xsltGenericDebugContext,
+                             "apply attribute set %s\n", attrib);
 #endif
-	    ncname = xmlSplitQName2(attrib, &prefix);
-	    if (ncname == NULL) {
-		ncname = attrib;
-		attrib = NULL;
-		prefix = NULL;
-	    }
+            ncname = xmlSplitQName2(attrib, &prefix);
+            if (ncname == NULL) {
+                ncname = attrib;
+                attrib = NULL;
+                prefix = NULL;
+            }
 
-	    style = ctxt->style;
-	    while (style != NULL) {
-		values = xmlHashLookup2(style->attributeSets, ncname, prefix);
-		while (values != NULL) {
-		    if (values->attr != NULL) {
-			xsltAttributeInternal(ctxt, node, values->attr,
-					      values->attr->_private, 1);
-		    }
-		    values = values->next;
-		}
-		style = xsltNextImport(style);
-	    }
-	    if (attrib != NULL)
-		xmlFree(attrib);
-	    if (ncname != NULL)
-		xmlFree(ncname);
-	    if (prefix != NULL)
-		xmlFree(prefix);
-	}
-	attrib = endattr;
+            style = ctxt->style;
+#ifdef WITH_DEBUGGER
+            if (style && (xslDebugStatus != XSLT_DEBUG_NONE)) {
+                values =
+                    xmlHashLookup2(style->attributeSets, ncname, prefix);
+                if (values)
+                    xslHandleDebugger(values->attr->parent, node, NULL,
+                                      ctxt);
+            }
+#endif
+            while (style != NULL) {
+                values =
+                    xmlHashLookup2(style->attributeSets, ncname, prefix);
+                while (values != NULL) {
+                    if (values->attr != NULL) {
+                        xsltAttributeInternal(ctxt, node, values->attr,
+                                              values->attr->_private, 1);
+                    }
+                    values = values->next;
+                }
+                style = xsltNextImport(style);
+            }
+            if (attrib != NULL)
+                xmlFree(attrib);
+            if (ncname != NULL)
+                xmlFree(ncname);
+            if (prefix != NULL)
+                xmlFree(prefix);
+        }
+        attrib = endattr;
     }
 }
 
